package com.xebialabs.deployit.cli.api;

import java.util.List;
import java.util.stream.Collectors;

import com.xebialabs.deployit.engine.api.dto.ActiveTaskStatus;
import com.xebialabs.deployit.engine.api.execution.*;
import org.joda.time.DateTime;

import com.xebialabs.deployit.cli.CliObject;
import com.xebialabs.deployit.cli.help.ClassHelp;
import com.xebialabs.deployit.cli.help.MethodHelp;
import com.xebialabs.deployit.cli.help.ParameterHelp;
import com.xebialabs.deployit.engine.api.dto.Paging;

import static com.google.common.collect.Lists.newArrayList;
import static com.xebialabs.deployit.cli.help.DateHelp.toLocalDate;

/**
 * Exposes the {@link com.xebialabs.deployit.engine.api.TaskBlockService}.
 *
 * @since Deployit 3.8
 */
@CliObject(name = "task2")
@ClassHelp(description = "Access to the task block engine of XL Deploy.")
public class TaskBlockClient extends DocumentedObject {

    private ProxiesInstance proxies;

    public TaskBlockClient() {
    }

    public TaskBlockClient(ProxiesInstance proxies) {
        this.proxies = proxies;
    }

    //
    // Method delegates to TaskService
    //
    @MethodHelp(description =
            "Lists the active tasks of the logged in user as summaries (without block information).",
            returns = "A list of tasks with blocks but without step information")
    public List<TaskWithBlock> getMyCurrentTaskSummaries() {
        return proxies.getTaskBlockRegistry().getMyCurrentTasks(FetchMode.SUMMARY);
    }

    @MethodHelp(description =
            "Lists the active tasks of the logged in user." +
                "Includes block information, prefer myCurrentTaskSummaries for better performance.",
            returns = "A list of tasks with blocks but without step information")
    public List<TaskWithBlock> getMyCurrentTasks() {
        return proxies.getTaskBlockRegistry().getMyCurrentTasks(FetchMode.FULL);
    }

    @MethodHelp(description =
            "Lists the active tasks of all users as summaries (without block information). Requires admin permission.",
            returns = "A list of tasks with blocks but without step information")
    public List<TaskWithBlock> getAllCurrentTaskSummaries() {
        return proxies.getTaskBlockRegistry().getAllCurrentTasks(FetchMode.SUMMARY);
    }

    @MethodHelp(description =
            "Lists the active tasks of the logged in user." +
                "Includes block information, prefer allCurrentTaskSummaries for better performance.",
            returns = "A list of tasks with blocks but without step information")
    public List<TaskWithBlock> getAllCurrentTasks() {
        return proxies.getTaskBlockRegistry().getAllCurrentTasks(FetchMode.FULL);
    }

    @MethodHelp(description =
            "Gets the task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")},
            returns = "The task with blocks but without step information.")
    public TaskWithBlock get(String taskId) {
        return proxies.getTaskBlockRegistry().getTask(taskId);
    }

    @MethodHelp(description =
            "Gets the task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "blockId", description = "The id of the block.")},
            returns = "The task with blocks but without step information.")
    public BlockState block(String taskId, String blockId) {
        return proxies.getTaskBlockRegistry().getBlock(taskId, blockId);
    }

    @MethodHelp(description =
            "Gets the task with steps.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "blockId", description = "The id of the block.")},
            returns = "The task, containing step information.")
    public StepBlockState steps(String taskId, String blockId) {
        return proxies.getTaskBlockRegistry().getSteps(taskId, blockId);
    }

    @MethodHelp(description =
            "Gets information about a step.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "stepId", description = "The position of the step in the step list.")},
            returns = "The step info.")
    public StepState step(String taskId, String stepId) {
        return proxies.getTaskBlockRegistry().getStep(taskId, stepId, null);
    }

    @MethodHelp(description =
            "Starts a task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")})
    public void start(String taskId) {
        proxies.getTaskBlockRegistry().start(taskId);
    }

    @MethodHelp(description = "Schedule a task for execution in the (near) future.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "dateTime", description = "The date and time of scheduled execution.")})
    public void schedule(String taskId, DateTime dateTime) {
        proxies.getTaskBlockRegistry().schedule(taskId, dateTime);
    }

    @MethodHelp(description =
            "Gracefully stops an active task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")})
    public void stop(String taskId) {
        proxies.getTaskBlockRegistry().stop(taskId);
    }

    @MethodHelp(description =
            "Aborts an active task", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")})
    public void abort(String taskId) {
        proxies.getTaskBlockRegistry().abort(taskId);
    }

    @MethodHelp(description =
            "Cancels a stopped task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")})
    public void cancel(String taskId) {
        proxies.getTaskBlockRegistry().cancel(taskId);
    }

    @MethodHelp(description =
            "Forcibly cancels a task. Use at your own peril! Clean-up phases will be executed on a best-effort basis, where already-failed steps are skipped, pause steps are ignored, and execution will continue after failure. Stop and Abort commands should still work. Use this only as a last resort for stuck tasks where normal cancel fails, since it may introduce discrepancies between reality and XL Deploy's view of it.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")})
    public void forceCancel(String taskId) {
        proxies.getTaskBlockRegistry().forceCancel(taskId);
    }

    @MethodHelp(description =
            "Archives an executed task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task.")})
    public void archive(String taskId) {
        proxies.getTaskBlockRegistry().archive(taskId);
    }

    @MethodHelp(description =
            "Skips steps of a task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "stepIds", description = "The ids of the steps to skip.")},
            returns = "The updated task with block information.")
    public TaskWithBlock skip(String taskId, List<String> stepIds) {
        return proxies.getTaskBlockRegistry().skip(taskId, newArrayList(stepIds));
    }

    @MethodHelp(description =
            "Unskips steps of a task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "stepIds", description = "The ids of the steps to skip.")},
            returns = "The updated task with block information.")
    public TaskWithBlock unskip(String taskId, List<String> stepIds) {
        return proxies.getTaskBlockRegistry().unskip(taskId, newArrayList(stepIds));
    }

    @MethodHelp(description =
            "Add a pause step to a task.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "stepPath", description = "The step path where the new pause step should appear.")},
            returns = "The step block containing the new pause step."
    )
    public StepBlockState addPause(String taskId, String stepPath) {
        return proxies.getTaskBlockRegistry().addPause(taskId, stepPath);
    }

    @MethodHelp(description =
            "Assigns a pending task to another principal.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "owner", description = "The new task owner.")},
            returns = "The task without step information.")
    public TaskWithBlock assign(String taskId, String owner) {
        return proxies.getTaskBlockRegistry().assign(taskId, owner);
    }

    @MethodHelp(description =
            "Takeover the ownership of a task from the current owner.", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "currentOwner", description = "The current task owner.")},
            returns = "The task without step information.")
    public TaskWithBlock takeover(String taskId, String currentOwner) {
        return proxies.getTaskBlockRegistry().takeover(taskId, currentOwner);
    }

    @MethodHelp(description =
            "Gets all task information from the repository's archive in the specified date range. At most 1000 results (unless a different limit was configured)", parameters = {
            @ParameterHelp(name = "beginDate", description = "Begin date from which to return tasks in 'MM/dd/yyyy' format."),
            @ParameterHelp(name = "endDate", description = "End date from which to return tasks in 'MM/dd/yyyy' format.")},
            returns = "A list of tasks without step information.")
    public List<TaskWithBlock> query(String beginDate, String endDate) {
        return query(beginDate, endDate, 1, -1);
    }


    @MethodHelp(description =
            "Gets one page of task information from the repository's archive in the specified date range.", parameters = {
            @ParameterHelp(name = "beginDate", description = "Begin date from which to return tasks in 'MM/dd/yyyy' format."),
            @ParameterHelp(name = "endDate", description = "End date from which to return tasks in 'MM/dd/yyyy' format."),
            @ParameterHelp(name = "page", description = "Which page to retrieve"),
            @ParameterHelp(name = "resultsPerPage", description = "How many results per page. Max 1000 (unless a different limit was configured)")},
            returns = "A list of tasks without step information.")
    public List<TaskWithBlock> query(String beginDate, String endDate, int page, int resultsPerPage) {
        return proxies.getTaskBlockRegistry().query(toLocalDate(beginDate), toLocalDate(endDate).plusDays(1), new Paging(page, resultsPerPage)).collect(Collectors.toList());
    }

    @MethodHelp(description = "Gets all task information, including steps, from the repository's archive in the specified date range. At most 1000 results (unless a different limit was configured)", parameters = {
            @ParameterHelp(name = "beginDate", description = "Begin date from which to return tasks in 'MM/dd/yyyy' format."),
            @ParameterHelp(name = "endDate", description = "End date from which to return tasks in 'MM/dd/yyyy' format.")},
            returns = "A list of tasks with their enclosed blocks and steps.")
    public List<TaskWithBlock> export(String beginDate, String endDate) {
        return export(beginDate, endDate, 1, -1);
    }

    @MethodHelp(description = "Gets one page task information, including steps, from the repository's archive in the specified date range.", parameters = {
            @ParameterHelp(name = "beginDate", description = "Begin date from which to return tasks in 'MM/dd/yyyy' format."),
            @ParameterHelp(name = "endDate", description = "End date from which to return tasks in 'MM/dd/yyyy' format."),
            @ParameterHelp(name = "page", description = "Which page to retrieve"),
            @ParameterHelp(name = "resultsPerPage", description = "How many results per page. Max 1000 (unless a different limit was configured)")},
            returns = "A list of tasks with their enclosed blocks and steps.")
    public List<TaskWithBlock> export(String beginDate, String endDate, int page, int resultsPerPage) {
        return proxies.getTaskBlockRegistry().export(toLocalDate(beginDate), toLocalDate(endDate).plusDays(1), new Paging(page, resultsPerPage)).collect(Collectors.toList());
    }

    @MethodHelp(description = "Gets the status of step or block by path", parameters = {
            @ParameterHelp(name = "taskId", description = "The id of the task."),
            @ParameterHelp(name = "path", description = "The path of the step/block.")},
            returns = "Status of the task step/block")
    public ActiveTaskStatus getStatus(String taskId, String path) {
        return proxies.getTaskBlockRegistry().getStatus(taskId, path);
    }

}
